// We need to process MIME types, which requires
// patform specific code, unforunately

#if defined(__linux__)
    #include <magic.h>
#endif

#include <QUrl>
#include <QFile>
#include <QDir>
#include <QProcess>
#include <QSettings>

#include "CHTTPServer.h"

CHTTPServer::CHTTPServer()
{
}

CHTTPServer::~CHTTPServer()
{
}

void CHTTPServer::run()
{
    // Emit a signal indicating that we're starting
    // the server up.

    emit statusString("Starting server...");

    // Find the settings and load them
    QSettings settings;

    m_document_root = settings.value("general/document_root","/var/www").toString();
    m_port          = settings.value("general/port",8081).toInt();

    QTcpServer server;
    m_server = &server;

    // listen for incomming connections...

    connect(&server,SIGNAL(newConnection()),this,SLOT(OnConnect()),Qt::DirectConnection);

    if(!server.listen(QHostAddress::Any,m_port))
    {
        emit statusString(tr("Error: port %1 could not be opened.").arg(m_port));
        return;
    }

    // Emit a signal indicating that the server started
    // up successfully.

    emit statusStarted();

    exec();

    // Emit the signal indicating that we're stopping
    // the server

    emit statusString("Stopping server...");

    // Stop the server

    server.close();

    // Now the server should be stopped

    emit statusStopped();
}

void CHTTPServer::Abort()
{
    if(!isRunning())
        return;

    quit();
}

void CHTTPServer::OnConnect()
{
    QTcpSocket * pSocket = m_server->nextPendingConnection();
    QHostAddress addr = pSocket->peerAddress();

    QString ip_addr = addr.toString();

    emit statusString(tr("Incoming request from %1").arg(ip_addr));

    QString request;

    while(pSocket->waitForReadyRead())
    {
        // Read as much as we can.
        QByteArray data = pSocket->readAll();

        request.append(data);

        // If the data contains \r\n\r\n, we're done!

        if(request.contains("\r\n\r\n"))
            break;
    }

    // Now parse the data

    ParseRequest(pSocket,request);

    // Wait for data writing to complete

    pSocket->waitForBytesWritten();

    delete pSocket;
}

void CHTTPServer::ParseRequest(QTcpSocket * pSocket,QString request)
{
    QStringList raw_headers = request.split("\r\n");

    // Make sure we got what we were looking for...

    if(!raw_headers.size())
    {
        SendError(pSocket,"400 Bad Request");

        emit statusString(tr("Invalid request received."));

        return;
    }

    // Now split that one too!

    QStringList status = raw_headers.at(0).split(" ");

    if(status.size() < 2)
    {
        SendError(pSocket,"400 Bad Request");

        emit statusString(tr("Invalid request received."));

        return;
    }

    // We need to figure out what page was requested

    // We start by putting the www path into the URL
    QUrl base,url;
    url.setUrl(tr("%1%2").arg(m_document_root).arg(status.at(1)));

    QUrl resolved = base.resolved(url);
    QString path = resolved.path();

    if(path.endsWith('/'))
    {
        // See if an index file exists
        QFile temp(path + "index.html");
        if(temp.exists())
            path.append("index.html");
    }

    emit statusString(tr("Page '%1' requested...").arg(path));

    // See if the filename ends with PHP

    if(path.contains('.'))
    {
        QString temp = path;
        temp.remove(0,path.lastIndexOf('.') + 1);

        if(!temp.compare("php",Qt::CaseInsensitive))
        {
            // We have a PHP script...
            // Now we need to attempt to launch the
            // PHP executable with our request data as STDIN

            qint64 size = WritePHPContent(pSocket,resolved,request);

            emit statusInfo(size);

            return;
        }
    }

    qint64 size = WriteStaticContent(pSocket,path,status.at(1));

    // emit a signal indicating that we processed this...

    emit statusInfo(size);
}

void CHTTPServer::SendError(QTcpSocket * pSocket,QString error)
{
    // We need to start by assembling the headers
    // for the error

    QList< QPair< QString,QString > > headers;

    headers.append(QPair< QString,QString >("Status",error));
    headers.append(QPair< QString,QString >("Content-Type","text/html"));

    // Response body

    QString error_str = tr("<h1>%1</h1><p>%2</p><hr><i>%3</i>")
                        .arg(error)
                        .arg("There was an error processing your request.")
                        .arg("JetHTTP 0.1 Beta");

    // We can use this to calculate the content-length header

    headers.append(QPair< QString,QString >("Content-Length",tr("%1").arg(error_str.length())));

    // Now output the header

    WriteHeader(pSocket,headers);

    // and the body

    pSocket->write(error_str.toUtf8());

    // Since we got an error, we need to emit a signal
    // so that the GUI can get updated.

    emit statusError();
}

void CHTTPServer::SendRedirect(QTcpSocket * pSocket, QString new_location)
{
    // Start by assembling the headers
    QList< QPair< QString,QString > > headers;

    headers.append(QPair< QString,QString >("Status","301 Moved Permanently"));
    headers.append(QPair< QString,QString >("Content-Type","text/html"));
    headers.append(QPair< QString,QString >("Location",new_location));
    headers.append(QPair< QString,QString >("Content-Length","0"));

    WriteHeader(pSocket,headers);
}

void CHTTPServer::WriteHeader(QTcpSocket * pSocket,QList< QPair< QString,QString > > headers)
{
    // Start by adding 'global' headers
    // to the list

    headers.append(QPair< QString,QString >("Server","JetHTTP 0.1 Beta [Qt]"));

    // We have the headers in a list.
    // We need to enumerate through them
    // and make sure none of them is the
    // 'Status:' header - we need that for
    // the first line.

    QString statusCode = "200 OK";
    QString header_str;

    QPair< QString,QString > header_line;

    foreach(header_line,headers)
    {
        if(header_line.first.toLower() == "status")
        {
            statusCode = header_line.second;
        }
        else
            header_str.append(tr("%1: %2\r\n").arg(header_line.first).arg(header_line.second));
    }

    // Now output the headers to the file

    pSocket->write(tr("HTTP/1.1 %1\r\n").arg(statusCode).toUtf8());
    pSocket->write(header_str.toUtf8());
    pSocket->write("\r\n");
}

QString CHTTPServer::GetMIMEType(QString filename)
{
    // We need to do some over-riding to make this
    // work. The first step is to check the extension

    QStringList parts = filename.split(".");

    if(parts.size())
    {
        if(parts.at(parts.size() - 1) == "js")
            return "application/x-javascript";
        if(parts.at(parts.size() - 1) == "css")
            return "text/css";
    }

    #if defined(__linux__)

    struct magic_set * magic = magic_open(MAGIC_MIME | MAGIC_CHECK);
    magic_load(magic,NULL);

    // Now return the result

    return magic_file(magic,filename.toUtf8().data());

    #endif
}

qint64 CHTTPServer::WriteStaticContent(QTcpSocket * pSocket,QString path,QString request_path)
{
    // Get the MIME type of the file in question

    QString mimeType = GetMIMEType(path);

    // Open the file for writing

    QFile file(path);
    file.open(QIODevice::ReadOnly);

    // Make sure we could open the file

    if(!file.isOpen())
    {
        QDir dir(path);

        if(!dir.exists())
        {
            SendError(pSocket,"404 File Not Found");
            return 0;
        }

        // It is a directory
        // now display its contents

        // make sure that it ends with '/'
        if(!path.endsWith('/'))
        {
            SendRedirect(pSocket,request_path + "/");
            return 0;
        }

        return WriteDirectoryContent(pSocket,path);
    }

    // Calculate the file's size and read
    // that many bytes from the file

    qint64 size = file.size();

    // Construct the headers necessary to return
    // to the client

    QList< QPair< QString,QString > > headers;

    headers.append(QPair< QString,QString >("Content-Type",mimeType));
    headers.append(QPair< QString,QString >("Content-Length",tr("%1").arg(size)));

    WriteHeader(pSocket,headers);

    // In order to do this properly,
    // we need to slowly feed the data to the
    // socket.

    qint64 bytes_so_far = 0;

    while(bytes_so_far < size)
    {
        qint64 bytes_to_read = 8192;

        if(bytes_so_far + bytes_to_read > size)
            bytes_to_read = size - bytes_so_far;

        QByteArray bytes = file.read(bytes_to_read);

        pSocket->write(bytes);
        pSocket->waitForBytesWritten();

        bytes_so_far += bytes_to_read;
    }

    return size;
}

qint64 CHTTPServer::WritePHPContent(QTcpSocket * pSocket, QUrl url, QString request)
{
    // We need to launch the PHP CLI and capture the output

    QProcess process;

    QProcessEnvironment env;

    // Environment variables required by CGI apps

    env.insert("SERVER_SOFTWARE",   "JetHTTP 0.1 Beta");
    env.insert("GATEWAY_INTERFACE", "CGI/1.1");
    env.insert("SERVER_PROTOCOL",   "HTTP/1.1");

    env.insert("QUERY_STRING",      url.encodedQuery());

    env.insert("REQUEST_URI",       url.toString());
    env.insert("SCRIPT_FILENAME",   url.path());

    env.insert("REQUEST_METHOD",    "GET");
    env.insert("REMOTE_ADDR",       pSocket->peerAddress().toString());

    env.insert("REDIRECT_STATUS",   "true");

    process.setProcessEnvironment(env);
    process.setProcessChannelMode(QProcess::MergedChannels);

    process.start(tr("php-cgi %1").arg(url.path()),QProcess::ReadWrite);

    process.waitForFinished();

    QString php_content = process.readAll();

    // We need to split the data into headers
    // returned and body returned...

    QStringList php_parts = php_content.split("\r\n\r\n");

    if(php_parts.size() < 2)
    {
        SendError(pSocket,"500 Internal Server Error");

        return 0;
    }

    // Break the first section of 'php_parts'
    // into a QStringList

    QStringList header_lines = php_parts.at(0).split("\r\n");

    // Now we need to loop through and add these to
    // our standard headers

    QList< QPair< QString,QString > > headers;

    headers.append(QPair< QString,QString >("Content-Length",tr("%1").arg(php_parts.at(1).length())));

    for(int i=0;i<header_lines.size();++i)
    {
        QStringList line_parts = header_lines.at(i).split(":");

        headers.append(QPair< QString,QString >(line_parts.at(0),line_parts.at(1)));
    }

    WriteHeader(pSocket,headers);

    qint64 size = php_parts.at(1).length();
    qint64 bytes_so_far = 0;

    while(bytes_so_far < size)
    {
        qint64 bytes_to_read = 8192;

        if(bytes_so_far + bytes_to_read > size)
            bytes_to_read = size - bytes_so_far;

        QByteArray bytes = php_parts.at(1).mid(bytes_so_far,bytes_to_read).toUtf8();

        pSocket->write(bytes);
        pSocket->waitForBytesWritten();

        bytes_so_far += bytes_to_read;
    }

    pSocket->write(php_parts.at(1).toUtf8());

    return php_parts.at(1).length();
}

qint64 CHTTPServer::WriteDirectoryContent(QTcpSocket * pSocket,QString path)
{
    // Begin by enumerating the contents of the directory
    QDir dir(path);

    QStringList contents = dir.entryList(QDir::NoFilter,QDir::Name);

    // First, some CSS:
    QString css;
    css.append("body { font-family: arial, sans, helvetica; } ");
    css.append("a { color: #555; text-decoration: none; } ");
    css.append("a:hover { color: #999; text-decoration: underline; } ");
    css.append("#footer { color: #666; float: right; margin-right: 10px; margin-top: 0px; } ");
    css.append(".dir { color: #22a; } ");

    // Now begin constructing an HTML page for this
    QString html;

    html.append(tr("<html><head><title>Contents of %1</title>").arg(path));
    html.append(tr("<style type='text/css'>%1</style>").arg(css));
    html.append(tr("</head><body><h2>Contents of <span class='dir'>%1</span></h2>").arg(path));
    html.append("<ul>");

    QString entry;
    foreach(entry,contents)
    {
        if(entry == "." || entry == "..")
            continue;

        html.append(tr("<li><a href='%1'>%1</a></li>").arg(entry));
    }

    html.append("</ul>");
    html.append("<hr><p id='footer'>Powered by JetHTTP</p>");
    html.append("</body></html>");

    QList< QPair< QString,QString > > headers;

    headers.append(QPair< QString,QString >("Content-Length",tr("%1").arg(html.length())));

    WriteHeader(pSocket,headers);

    pSocket->write(html.toUtf8());

    return 0;
}
